@@ -1,6 +1,5 @@
-from __future__ import annotations
 
-from typing import Dict, Any, List, Optional, Union, Callable, Tuple
+from typing import Dict, Any, List, Optional, Union, Callable, Tuple, Literal
 
 
 from dataclasses import dataclass
@@ -8,9 +7,11 @@ import warnings
 
 import numpy as np
 
-from .crossovers import Crossover
-from .mutations import Mutations
-from .selections import Selection
+from .aliases import array1D, array2D, TypeAlias
+
+from .crossovers import Crossover, CrossoverFunc
+from .mutations import Mutations, MutationIntFunc, MutationFloatFunc
+from .selections import Selection, SelectionFunc
 
 from .utils import can_be_prob, union_to_matrix
 
@@ -27,8 +28,6 @@ class DictLikeGetSet:
 
 
 
-
-
 _algorithm_params_slots = {
     'max_num_iteration',
     'max_iteration_without_improv',
@@ -44,6 +43,7 @@ _algorithm_params_slots = {
     'selection_type'
 }
 
+
 @dataclass
 class AlgorithmParams(DictLikeGetSet):
 
@@ -61,12 +61,27 @@ class AlgorithmParams(DictLikeGetSet):
     elit_ratio: float = 0.04
     parents_portion: float = 0.3
 
-    crossover_type: Union[str, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = 'uniform'
-    mutation_type: Union[str, Callable[[float, float, float], float]] = 'uniform_by_center'
-    mutation_discrete_type: Union[str, Callable[[int, int, int], int]] = 'uniform_discrete'
-    selection_type: Union[str, Callable[[np.ndarray, int], np.ndarray]] = 'roulette'
+    crossover_type: Union[str, CrossoverFunc] = 'uniform'
+    mutation_type: Union[str, MutationFloatFunc] = 'uniform_by_center'
+    mutation_discrete_type: Union[str, MutationIntFunc] = 'uniform_discrete'
+    selection_type: Union[str, SelectionFunc] = 'roulette'
 
-    def _check_if_valid(self):
+    __annotations__ = {
+        'max_num_iteration': Optional[int],
+        'max_iteration_without_improv': Optional[int],
+        'population_size': int,
+        'mutation_probability': float,
+        'mutation_discrete_probability': Optional[float],
+        'crossover_probability': Optional[float],
+        'elit_ratio': float,
+        'parents_portion': float,
+        'crossover_type': Union[str, CrossoverFunc],
+        'mutation_type': Union[str, MutationFloatFunc],
+        'mutation_discrete_type': Union[str, MutationIntFunc],
+        'selection_type': Union[str, SelectionFunc]
+    }
+
+    def check_if_valid(self) -> None:
 
         assert int(self.population_size) > 0, f"population size must be integer and >0, not {self.population_size}"
         assert (can_be_prob(self.parents_portion)), "parents_portion must be in range [0,1]"
@@ -78,27 +93,26 @@ class AlgorithmParams(DictLikeGetSet):
             self.max_iteration_without_improv = None
 
     def get_CMS_funcs(self) -> Tuple[
-        Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]],
-        Callable[[float, float, float], float],
-        Callable[[int, int, int], int],
-        Callable[[np.ndarray, int], np.ndarray]
+        CrossoverFunc,
+        MutationFloatFunc,
+        MutationIntFunc,
+        SelectionFunc
     ]:
         """
         returns gotten crossover, mutation, discrete mutation, selection
         as necessary functions
         """
 
-        result = []
+        result: List[Callable] = []
         for name, value, dct in (
             ('crossover', self.crossover_type, Crossover.crossovers_dict()),
             ('mutation', self.mutation_type, Mutations.mutations_dict()),
             ('mutation_discrete', self.mutation_discrete_type, Mutations.mutations_discrete_dict()),
             ('selection', self.selection_type, Selection.selections_dict())
         ):
-
             if type(value) == str:
                 if value not in dct:
-                    raise Exception(
+                    raise ValueError(
                         f"unknown name of {name}: '{value}', must be from {tuple(dct.keys())} or a custom function"
                     )
                 result.append(dct[value])
@@ -108,7 +122,6 @@ class AlgorithmParams(DictLikeGetSet):
 
         return tuple(result)
 
-
     @staticmethod
     def from_dict(dct: Dict[str, Any]):
 
@@ -122,12 +135,29 @@ class AlgorithmParams(DictLikeGetSet):
         return result
 
 
+GenerationConvertible: TypeAlias = Union[
+    'Generation',
+    str,
+    Dict[Literal['population', 'scores'], Union[array2D, array1D]],
+    array2D,
+    Tuple[
+        Optional[array2D],
+        Optional[array1D]
+    ]
+]
+
+
 @dataclass
 class Generation(DictLikeGetSet):
-    variables: Optional[np.ndarray] = None
-    scores: Optional[np.ndarray] = None
+    variables: Optional[array2D] = None
+    scores: Optional[array1D] = None
 
-    def __check_dims(self):
+    __annotations__ = {
+        'variables': Optional[array2D],
+        'scores': Optional[array1D]
+    }
+
+    def __check_dims(self) -> None:
         if self.variables is not None:
             assert len(self.variables.shape) == 2, f"'variables' must be matrix with shape (objects, dimensions), not {self.variables.shape}"
             if self.scores is not None:
@@ -135,21 +165,20 @@ class Generation(DictLikeGetSet):
                 assert self.variables.shape[0] == self.scores.size, f"count of objects ({self.variables.shape[0]}) must be equal to count of scores ({self.scores.size})"
 
     @property
-    def size(self):
+    def size(self) -> int:
         return self.scores.size
 
     @property
-    def dim_size(self):
+    def dim_size(self) -> int:
         return self.variables.shape[1]
 
-    def as_wide_matrix(self):
+    def as_wide_matrix(self) -> array2D:
         # should not be used in main code -- was needed for old versions
         return union_to_matrix(self.variables, self.scores)
 
     def save(self, path: str):
         np.savez(path, population=self.variables, scores=self.scores)
 
-
     @staticmethod
     def load(path: str):
         try:
@@ -163,20 +192,10 @@ class Generation(DictLikeGetSet):
 
         return Generation(variables=st['population'], scores=st['scores'])
 
-
     @staticmethod
     def from_object(
         dim: int,
-        object: Union[
-            str,
-            Dict[str, np.ndarray],
-            Generation,
-            np.ndarray,
-            Tuple[
-                Optional[np.ndarray],
-                Optional[np.ndarray]
-            ]
-        ]
+        object: GenerationConvertible
     ):
 
         obj_type = type(object)
@@ -202,12 +221,12 @@ class Generation(DictLikeGetSet):
 
             generation = Generation(variables=variables, scores=scores)
 
-
         elif obj_type == dict:
-            assert (('variables' in object and 'scores' in object) and (
-                             object['variables'] is None or object['scores'] is None) or (
-                                     object['variables'].shape[0] == object[
-                                 'scores'].size)), "start_generation object must contain 'variables' and 'scores' keys which are None or 2D- and 1D-arrays with same shape"
+            assert (
+                ('variables' in object and 'scores' in object) and
+                (object['variables'] is None or object['scores'] is None) or
+                (object['variables'].shape[0] == object['scores'].size)
+            ), "start_generation object must contain 'variables' and 'scores' keys which are None or 2D- and 1D-arrays with same shape"
 
             generation = Generation(variables=object['variables'], scores=object['scores'])
 
@@ -216,7 +235,6 @@ class Generation(DictLikeGetSet):
         else:
             raise TypeError(f"invalid type of generation! Must be in (Union[str, Dict[str, np.ndarray], Generation, np.ndarray, Tuple[Optional[np.ndarray], Optional[np.ndarray]]]), not {obj_type}")
 
-
         generation.__check_dims()
 
         if generation.variables is not None:
@@ -225,7 +243,7 @@ class Generation(DictLikeGetSet):
         return generation
 
     @staticmethod
-    def from_pop_matrix(pop: np.ndarray):
+    def from_pop_matrix(pop: array2D):
         warnings.warn("depricated! pop matrix style will be removed at version 7, use samples and scores separetly")
         return Generation(
             variables=pop[:, :-1],
@@ -238,12 +256,18 @@ class GAResult(DictLikeGetSet):
 
     last_generation: Generation
 
+    __annotations__ = {
+        'last_generation': Generation
+    }
+
     @property
-    def variable(self):
+    def variable(self) -> array1D:
         return self.last_generation.variables[0]
+
     @property
-    def score(self):
+    def score(self) -> float:
         return self.last_generation.scores[0]
+
     @property
     def function(self):
         warnings.warn(f"'function' field is deprecated, will be removed in version 7, use 'score' to get best population score")
@@ -266,10 +290,10 @@ class MiddleCallbackData(DictLikeGetSet):
     mutation_prob: float
     mutation_discrete_prob: float
 
-    mutation: Callable[[float, float, float], float]
-    mutation_discrete: Callable[[int, int, int], int]
-    crossover: Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]
-    selection: Callable[[np.ndarray, int], np.ndarray]
+    mutation: MutationFloatFunc
+    mutation_discrete: MutationIntFunc
+    crossover: CrossoverFunc
+    selection: SelectionFunc
 
     current_stagnation: int
     max_stagnation: int
@@ -277,7 +301,27 @@ class MiddleCallbackData(DictLikeGetSet):
     parents_portion: float
     elit_ratio: float
 
-    set_function: Callable[[np.ndarray], np.ndarray]
+    set_function: Callable[[array2D], array1D]
+    
+    __annotations__ = {
+        'reason_to_stop': Optional[str],
+        'last_generation': Generation,
+        'current_generation': int,
+        'report_list': List[float],
+        'mutation_prob': float,
+        'mutation_discrete_prob': float,
+        'mutation': MutationFloatFunc,
+        'mutation_discrete': MutationIntFunc,
+        'crossover': CrossoverFunc,
+        'selection': SelectionFunc,
+        'current_stagnation': int,
+        'max_stagnation': int,
+        'parents_portion': float,
+        'elit_ratio': float,
+        'set_function': Callable[[array2D], array1D]
+    }
+
+
 
 
 
