@@ -1,17 +1,22 @@
-import math
+
+from typing import Callable, Tuple, Dict
+
 import random
 import numpy as np
 
+from .aliases import TypeAlias, array1D
+
+CrossoverFunc: TypeAlias = Callable[[array1D, array1D], Tuple[array1D, array1D]]
 
 
-def get_copies(x: np.ndarray, y: np.ndarray):
+def get_copies(x: array1D, y: array1D) -> Tuple[array1D, array1D]:
     return x.copy(), y.copy()
 
 
 class Crossover:
 
     @staticmethod
-    def crossovers_dict():
+    def crossovers_dict() -> Dict[str, CrossoverFunc]:
         return {
             'one_point': Crossover.one_point(),
             'two_point': Crossover.two_point(),
@@ -21,12 +26,12 @@ class Crossover:
         }
     
     @staticmethod
-    def one_point():
+    def one_point() -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             ofs1, ofs2 = get_copies(x, y)
         
-            ran=np.random.randint(0, x.size)
+            ran = np.random.randint(0, x.size)
             
             ofs1[:ran] = y[:ran]
             ofs2[:ran] = x[:ran]
@@ -35,25 +40,24 @@ class Crossover:
         return func
     
     @staticmethod
-    def two_point():
+    def two_point() -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             ofs1, ofs2 = get_copies(x, y)
         
-            ran1=np.random.randint(0, x.size)
-            ran2=np.random.randint(ran1, x.size)
+            ran1 = np.random.randint(0, x.size)
+            ran2 = np.random.randint(ran1, x.size)
             
             ofs1[ran1:ran2] = y[ran1:ran2]
             ofs2[ran1:ran2] = x[ran1:ran2]
-              
-            
+
             return ofs1, ofs2
         return func
     
     @staticmethod
-    def uniform():
+    def uniform() -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             ofs1, ofs2 = get_copies(x, y)
         
             ran = np.random.random(x.size) < 0.5
@@ -65,9 +69,9 @@ class Crossover:
         return func
     
     @staticmethod
-    def segment(prob: int = 0.6):
+    def segment(prob: int = 0.6) -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             
             ofs1, ofs2 = get_copies(x, y)
             
@@ -82,9 +86,9 @@ class Crossover:
         return func
     
     @staticmethod
-    def shuffle():
+    def shuffle() -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             
             ofs1, ofs2 = get_copies(x, y)
             
@@ -102,14 +106,14 @@ class Crossover:
         return func
     
     @staticmethod
-    def uniform_window(window: int = 7):
+    def uniform_window(window: int = 7) -> CrossoverFunc:
 
         base_uniform = Crossover.uniform()
 
         def func(x: np.ndarray, y: np.ndarray):
 
             if x.size % window != 0:
-                raise Exception(f"dimension {x.size} cannot be divided by window {window}")
+                raise ValueError(f"dimension {x.size} cannot be divided by window {window}")
             
             items = int(x.size/window)
 
@@ -127,13 +131,9 @@ class Crossover:
                     ofs1[sls] = y[sls]                    
 
             return ofs1, ofs2
-        
 
         return func
 
-
-
-
     #
     #
     # ONLY FOR REAL VARIABLES
@@ -141,9 +141,9 @@ class Crossover:
     #
     
     @staticmethod
-    def arithmetic():
+    def arithmetic() -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             b = random.random()
             a = 1-b
             return a*x + b*y, a*y + b*x
@@ -151,9 +151,9 @@ class Crossover:
         return func
     
     @staticmethod
-    def mixed(alpha: float = 0.5):
+    def mixed(alpha: float = 0.5) -> CrossoverFunc:
         
-        def func(x: np.ndarray, y: np.ndarray):
+        def func(x: array1D, y: array1D):
             
             a = np.empty(x.size)
             b = np.empty(y.size)
