@@ -1,5 +1,4 @@
-'''
-
+"""
 Copyright 2021 Demetry Pascal (forked from Ryan (Mohammad) Solgi)
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of 
@@ -19,8 +18,7 @@ THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 SOFTWARE.
-
-'''
+"""
 
 
 from typing import Callable, List, Tuple, Optional, Dict, Any, Union, Sequence, Set
@@ -42,30 +40,30 @@ from OppOpPopInit import init_population, SampleInitializers, OppositionOperator
 
 #region INTERNAL IMPORTS
 
-from .classes import AlgorithmParams, Generation, MiddleCallbackData, GAResult
+from .aliases import array1D, array2D
+
+from .classes import AlgorithmParams, Generation, MiddleCallbackData, GAResult, GenerationConvertible
 
 from .initializer import Population_initializer
 from .plotting_tools import plot_pop_scores, plot_several_lines
 
-from .utils import (can_be_prob, is_numpy, is_current_gen_number,
-                    union_to_matrix,
-                    fast_max, fast_min)
+from .utils import can_be_prob, is_numpy, is_current_gen_number, fast_min, random_indexes_pair
+
+from .callbacks import MiddleCallbackFunc, CallbackFunc
 
 #endregion
 
 
-
-
 class geneticalgorithm2:
     
-    '''
+    """
     Genetic Algorithm (Elitist version) for Python
     
     An implementation of elitist genetic algorithm for solving problems with
     continuous, integers, or mixed variables.
     
     repo path https://github.com/PasaOpasen/geneticalgorithm2
-    '''
+    """
     
     default_params = AlgorithmParams()
     PROGRESS_BAR_LEN = 20
@@ -78,25 +76,25 @@ class geneticalgorithm2:
         return self.result
 
     @property
-    def needs_mutation(self):
+    def needs_mutation(self) -> bool:
         return self.prob_mut > 0 or self.prob_mut_discrete > 0
 
     #region INIT
 
     def __init__(
         self,
-        function: Callable[[np.ndarray], float],
+        function: Callable[[array1D], float],
 
         dimension: int,
         variable_type: Union[str, Sequence[str]] = 'bool',
-        variable_boundaries: Optional[Union[np.ndarray, Sequence[Tuple[float, float]]]] = None,
+        variable_boundaries: Optional[Union[array2D, Sequence[Tuple[float, float]]]] = None,
 
-        variable_type_mixed = None,
+        variable_type_mixed=None,
 
         function_timeout: float = 10,
         algorithm_parameters: Union[AlgorithmParams, Dict[str, Any]] = default_params
     ):
-        '''
+        """
         @param function <Callable[[np.ndarray], float]> - the given objective function to be minimized
         NOTE: This implementation minimizes the given objective function.
         (For maximization multiply function by a negative sign: the absolute
@@ -138,7 +136,7 @@ class geneticalgorithm2:
         for more details and examples of implementation please visit:
             https://github.com/PasaOpasen/geneticalgorithm2
   
-        '''
+        """
 
 
         # all default fields
@@ -149,7 +147,7 @@ class geneticalgorithm2:
         # self.discrete_mutation: Callable[[int, int, int], int] = None
         # self.selection: Callable[[np.ndarray, int], np.ndarray] = None
 
-        self.f: Callable[[np.ndarray], float] = None
+        self.f: Callable[[array1D], float] = None
         self.funtimeout: float = None
         self.set_function: Callable[[np.ndarray], np.ndarray] = None
 
@@ -158,7 +156,7 @@ class geneticalgorithm2:
         # self.indexes_int: np.ndarray = None
         # self.indexes_float: np.ndarray = None
 
-        self.checked_reports: List[Tuple[str, Callable[[np.ndarray], None]]] = None
+        self.checked_reports: List[Tuple[str, Callable[[array1D], None]]] = None
 
         self.population_size: int = None
         self.progress_stream = None
@@ -170,7 +168,7 @@ class geneticalgorithm2:
         if type(algorithm_parameters) != AlgorithmParams:
             algorithm_parameters = AlgorithmParams.from_dict(algorithm_parameters)
 
-        algorithm_parameters._check_if_valid()
+        algorithm_parameters.check_if_valid()
         self.param = algorithm_parameters
         self.crossover, self.real_mutation, self.discrete_mutation, self.selection = algorithm_parameters.get_CMS_funcs()
 
@@ -218,6 +216,9 @@ class geneticalgorithm2:
 
         self._set_report()
 
+        # specify this function to speed up or change default behaviour
+        self.fill_children: Optional[Callable[[array2D, int], None]] = None
+        """custom function which adds children for population POP where POP[:parents_count] are parents lines and next lines are for children"""
 
     def _set_types_indexes(self, variable_type: Union[str, Sequence[str]]):
 
@@ -225,7 +226,7 @@ class geneticalgorithm2:
         self.indexes_int = np.array([])
         self.indexes_float = np.array([])
 
-        VALID_STRINGS = ( 'bool', 'int', 'real')
+        VALID_STRINGS = ('bool', 'int', 'real')
         assert_message = f"\n variable_type must be 'bool', 'int', 'real' or a sequence with 'int' and 'real', got {variable_type}"
 
         if type(variable_type) == str:
@@ -278,7 +279,6 @@ class geneticalgorithm2:
         assert elit_ratio >= 0
         self.elit_count = elit_ratio if type(elit_ratio) == int else math.ceil(pop_size*elit_ratio)
 
-
     def _set_max_iterations(self):
 
         if self.param.max_num_iteration is None:
@@ -333,7 +333,7 @@ class geneticalgorithm2:
         for name, _ in self.checked_reports:
             setattr(self, name, [])
 
-    def _update_report(self, scores: np.ndarray):
+    def _update_report(self, scores: array1D):
         """
         append report value to the end of field
         """
@@ -363,7 +363,7 @@ class geneticalgorithm2:
     def __repr__(self):
         return self.__str__()
 
-    def _simulate(self, sample: np.ndarray):
+    def _simulate(self, sample: array1D):
 
         obj = None
         eval_time = time.time()
@@ -379,8 +379,9 @@ class geneticalgorithm2:
         assert obj is not None, f"the given function was running like {eval_time} seconds and does not provide any output"
 
         tp = type(obj)
-        assert (tp in (int, float) or np.issubdtype(tp, np.floating) or np.issubdtype(tp,
-                                                                                      np.integer)), f"Minimized function should return a number, but got '{obj}' object with type {tp}"
+        assert (
+            tp in (int, float) or np.issubdtype(tp, np.floating) or np.issubdtype(tp, np.integer)
+        ), f"Minimized function should return a number, but got '{obj}' object with type {tp}"
 
         return obj, eval_time
 
@@ -407,27 +408,27 @@ class geneticalgorithm2:
         # deprecated
         disable_progress_bar: bool = False,
 
-        set_function: Optional[Callable[[np.ndarray], np.ndarray]] = None,
+        set_function: Optional[Callable[[array2D], array1D]] = None,
         apply_function_to_parents: bool = False,
-        start_generation: Union[str, Dict[str, np.ndarray], Generation, np.ndarray, Tuple[Optional[np.ndarray], Optional[np.ndarray]]] = Generation(),
+        start_generation: GenerationConvertible = Generation(),
         studEA: bool = False,
         mutation_indexes: Optional[Union[Sequence[int], Set[int]]] = None,
 
-        init_creator: Optional[Callable[[], np.ndarray]] = None,
-        init_oppositors: Optional[Sequence[Callable[[np.ndarray], np.ndarray]]] = None,
+        init_creator: Optional[Callable[[], array1D]] = None,
+        init_oppositors: Optional[Sequence[Callable[[array1D], array1D]]] = None,
 
-        duplicates_oppositor: Optional[Callable[[np.ndarray], np.ndarray]] = None,
+        duplicates_oppositor: Optional[Callable[[array1D], array1D]] = None,
         remove_duplicates_generation_step: Optional[int] = None,
 
-        revolution_oppositor: Optional[Callable[[np.ndarray], np.ndarray]] = None,
+        revolution_oppositor: Optional[Callable[[array1D], array1D]] = None,
         revolution_after_stagnation_step: Optional[int] = None,
         revolution_part: float = 0.3,
 
-        population_initializer: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = Population_initializer(select_best_of = 1, local_optimization_step = 'never', local_optimizer = None),
+        population_initializer: Tuple[int, Callable[[array2D, array1D], Tuple[array2D, array1D]]] = Population_initializer(select_best_of=1, local_optimization_step='never', local_optimizer=None),
 
         stop_when_reached: Optional[float] = None,
-        callbacks: Optional[Sequence[Callable[[int, List[float],  np.ndarray, np.ndarray], None]]] = None,
-        middle_callbacks: Optional[Sequence] = None, #+
+        callbacks: Optional[Sequence[CallbackFunc]] = None,
+        middle_callbacks: Optional[Sequence[MiddleCallbackFunc]] = None, #+
         time_limit_secs: Optional[float] = None,
         save_last_generation_as: Optional[str] = None,
         seed: Optional[int] = None
@@ -436,8 +437,12 @@ class geneticalgorithm2:
         runs optimization process
 
         @param no_plot: do not plot results using matplotlib by default
+
+        @param disable_printing: do not print log info of optimization process
         
         @param progress_bar_stream: 'stdout', 'stderr' or None to disable progress bar
+
+        @param disable_progress_bar:
                 
         @param set_function : 2D-array -> 1D-array function, which applyes to matrix of population (size (samples, dimention))
         to estimate their values
@@ -479,7 +484,7 @@ class geneticalgorithm2:
             )
             progress_bar_stream = None
 
-        enable_printing = not disable_printing
+        enable_printing: bool = not disable_printing
 
         start_generation = Generation.from_object(self.dim, start_generation)
 
@@ -497,8 +502,12 @@ class geneticalgorithm2:
         # randomstate = np.random.default_rng(random.randint(0, 1000) if seed is None else seed)
         # self.randomstate = randomstate
 
-        show_progress = (lambda t, t2, s: self.__progress(t, t2, status=s)) if progress_bar_stream is not None else (lambda t, t2, s: None)
-        if progress_bar_stream is not None:
+        # using bool flag is faster than using empty function with generated string arguments
+        SHOW_PROGRESS = progress_bar_stream is not None
+        if SHOW_PROGRESS:
+
+            show_progress = lambda t, t2, s: self.__progress(t, t2, status=s)
+
             if progress_bar_stream == 'stdout':
                 self.progress_stream = sys.stdout
             elif progress_bar_stream == 'stderr':
@@ -507,13 +516,15 @@ class geneticalgorithm2:
                 raise Exception(
                     f"wrong value {progress_bar_stream} of progress_bar_stream, must be 'stdout'/'stderr'/None"
                 )
+        else:
+            show_progress = None
 
         stop_by_val = (lambda best_f: False) if stop_when_reached is None else (lambda best_f: best_f <= stop_when_reached)
 
-        t = 0
-        count_stagnation = 0
-        pop: np.ndarray = None
-        scores: np.ndarray = None
+        t: int = 0
+        count_stagnation: int = 0
+        pop: array2D = None
+        scores: array1D = None
 
         #
         #
@@ -582,8 +593,8 @@ class geneticalgorithm2:
             def total_callback(
                 generation_number: int,
                 report_list: List[float],
-                last_population: np.ndarray,
-                last_scores: np.ndarray
+                last_population: array2D,
+                last_scores: array1D
             ):
                 for cb in callbacks:
                     cb(generation_number, report_list, last_population, last_scores)
@@ -621,11 +632,11 @@ class geneticalgorithm2:
             self.creator = SampleInitializers.Combined(
                 minimums=[v[0] for v in self.var_bounds],
                 maximums=[v[1] for v in self.var_bounds],
-                indexes = (
+                indexes=(
                     self.indexes_int,
                     self.indexes_float
                 ),
-                creator_initializers = (
+                creator_initializers=(
                     SampleInitializers.RandomInteger,
                     SampleInitializers.Uniform
                 )
@@ -640,36 +651,39 @@ class geneticalgorithm2:
 
         # event for removing duplicates
         if remove_duplicates_generation_step is None:
-            def remover(pop: np.ndarray, scores: np.ndarray, gen: int) -> Tuple[
-                np.ndarray,
-                np.ndarray
+            def remover(pop: array2D, scores: array1D, gen: int) -> Tuple[
+                array2D,
+                array1D
             ]:
                 return pop, scores
         else:
             
-            def without_dup(pop: np.ndarray, scores: np.ndarray):  # returns population without dups
+            def without_dup(pop: array2D, scores: array1D):  # returns population without dups
                 _, index_of_dups = np.unique(pop, axis=0, return_index=True)
                 return pop[index_of_dups], scores[index_of_dups], scores.size - index_of_dups.size
             
             if self.dup_oppositor is None:  # if there is no dup_oppositor, use random creator
-                def remover(pop: np.ndarray, scores: np.ndarray, gen: int) -> Tuple[
-                    np.ndarray,
-                    np.ndarray
+                def remover(pop: array2D, scores: array1D, gen: int) -> Tuple[
+                    array2D,
+                    array1D
                 ]:
                     if gen % remove_duplicates_generation_step != 0:
                         return pop, scores
 
                     pp, sc, count_to_create = without_dup(pop, scores)  # pop without dups
                     
-                    if count_to_create == 0: 
-                        show_progress(t, self.max_iterations,
+                    if count_to_create == 0:
+                        if SHOW_PROGRESS:
+                            show_progress(t, self.max_iterations,
                                       f"GA is running...{t} gen from {self.max_iterations}. No dups!")
                         return pop, scores
 
                     pp2 = SampleInitializers.CreateSamples(self.creator, count_to_create)  # new pop elements
                     pp2_scores = self.set_function(pp2)  # new elements values
-                    
-                    show_progress(t, self.max_iterations, f"GA is running...{t} gen from {self.max_iterations}. Kill dups!")
+
+                    if SHOW_PROGRESS:
+                        show_progress(t, self.max_iterations,
+                                      f"GA is running...{t} gen from {self.max_iterations}. Kill dups!")
                     
                     new_pop = np.vstack((pp, pp2))
                     new_scores = np.concatenate((sc, pp2_scores))
@@ -689,8 +703,10 @@ class geneticalgorithm2:
 
                     pp, sc, count_to_create = without_dup(pop, scores)  # pop without dups
 
-                    if count_to_create == 0: 
-                        show_progress(t, self.max_iterations, f"GA is running...{t} gen from {self.max_iterations}. No dups!")
+                    if count_to_create == 0:
+                        if SHOW_PROGRESS:
+                            show_progress(t, self.max_iterations,
+                                          f"GA is running...{t} gen from {self.max_iterations}. No dups!")
                         return pop, scores
 
                     if count_to_create > sc.size:
@@ -700,8 +716,9 @@ class geneticalgorithm2:
                     pp2 = OppositionOperators.Reflect(pp[-count_to_create:], self.dup_oppositor)  # new pop elements
                     pp2_scores = self.set_function(pp2)  # new elements values
 
-                    show_progress(t, self.max_iterations,
-                                  f"GA is running...{t} gen from {self.max_iterations}. Kill dups!")
+                    if SHOW_PROGRESS:
+                        show_progress(t, self.max_iterations,
+                                      f"GA is running...{t} gen from {self.max_iterations}. Kill dups!")
 
                     new_pop = np.vstack((pp, pp2))
                     new_scores = np.concatenate((sc, pp2_scores))
@@ -711,9 +728,9 @@ class geneticalgorithm2:
 
         # event for revolution
         if revolution_after_stagnation_step is None:
-            def revolution(pop: np.ndarray, scores: np.ndarray, stagnation_count: int) -> Tuple[
-                np.ndarray,
-                np.ndarray
+            def revolution(pop: array2D, scores: array1D, stagnation_count: int) -> Tuple[
+                array2D,
+                array1D
             ]:
                 return pop, scores
         else:
@@ -723,9 +740,9 @@ class geneticalgorithm2:
                 )
             assert callable(revolution_oppositor)
             
-            def revolution(pop: np.ndarray, scores: np.ndarray, stagnation_count: int) -> Tuple[
-                np.ndarray,
-                np.ndarray
+            def revolution(pop: array2D, scores: array1D, stagnation_count: int) -> Tuple[
+                array2D,
+                array1D
             ]:
                 if stagnation_count < revolution_after_stagnation_step:
                     return pop, scores
@@ -737,8 +754,10 @@ class geneticalgorithm2:
                 combined = np.vstack((pop, pp2))
                 combined_scores = np.concatenate((scores, pp2_scores))
                 args = combined_scores.argsort()
-                
-                show_progress(t, self.max_iterations, f"GA is running...{t} gen from {self.max_iterations}. Revolution!")
+
+                if SHOW_PROGRESS:
+                    show_progress(t, self.max_iterations,
+                                  f"GA is running...{t} gen from {self.max_iterations}. Revolution!")
 
                 args = args[:scores.size]
                 return combined[args], combined_scores[args]
@@ -823,12 +842,16 @@ class geneticalgorithm2:
                 f"Best score before optimization: {self.best_function}"
             )
 
-        t = 1
-        count_stagnation = 0  # iterations without progress
+        t: int = 1
+        count_stagnation: int = 0
+        """iterations without progress"""
         reason_to_stop: Optional[str] = None
 
         # gets indexes of 2 parents to crossover
-        get_parents_inds = (lambda: (0, random.randrange(1, self.parents_count))) if studEA else (lambda: random.sample(range(self.parents_count), 2))
+        if studEA:
+            get_parents_inds = lambda: (0, random.randrange(1, self.parents_count))
+        else:
+            get_parents_inds = lambda: random_indexes_pair(self.parents_count)
 
         #  while no reason to stop
         while True:
@@ -843,7 +866,8 @@ class geneticalgorithm2:
                 reason_to_stop = f'time is done: {time.time() - start_time} >= {time_limit_secs}'
 
             if reason_to_stop is not None:
-                show_progress(t, self.max_iterations, f"GA is running... STOP! {reason_to_stop}")
+                if SHOW_PROGRESS:
+                    show_progress(t, self.max_iterations, f"GA is running... STOP! {reason_to_stop}")
                 break
 
             if scores[0] < self.best_function:  # if there is progress
@@ -852,52 +876,61 @@ class geneticalgorithm2:
             else:
                 count_stagnation += 1
 
-            show_progress(
-                t,
-                self.max_iterations,
-                f"GA is running...{t} gen from {self.max_iterations}...best value = {self.best_function}"
-            )
+            if SHOW_PROGRESS:
+                show_progress(
+                    t,
+                    self.max_iterations,
+                    f"GA is running...{t} gen from {self.max_iterations}...best value = {self.best_function}"
+                )
 
             # Select parents
             
-            par = np.empty((self.parents_count, self.dim))
-            par_scores = np.empty(self.parents_count)
-            
-            # elit parents
+            par: array2D = np.empty((self.parents_count, self.dim))
+            """samples chosen to create new samples"""
+            par_scores: array1D = np.empty(self.parents_count)
+
             elit_slice = slice(None, self.elit_count)
-            # copy needs because generation wil be removed after parents selection
+            """elit parents"""
+
+            # copy needs because the generation will be removed after parents selection
             par[elit_slice] = pop[elit_slice].copy()
             par_scores[elit_slice] = scores[elit_slice].copy()
-                
-            # non-elit parents indexes
+
             new_par_inds = self.selection(scores[self.elit_count:], self.parents_count - self.elit_count).astype(np.int16) + self.elit_count
+            """non-elit parents indexes"""
             #new_par_inds = self.selection(scores, self.parents_count - self.elit_count).astype(np.int16)
             par_slice = slice(self.elit_count, self.parents_count)
             par[par_slice] = pop[new_par_inds].copy()
             par_scores[par_slice] = scores[new_par_inds].copy()
 
-            # New generation
             pop = np.empty((self.population_size, self.dim))
+            """new generation"""
             scores = np.empty(self.population_size)
+            """new generation scores"""
 
             parents_slice = slice(None, self.parents_count)
-            pop[parents_slice] = par[parents_slice]
-            scores[parents_slice] = par_scores[parents_slice]
+            pop[parents_slice] = par
+            scores[parents_slice] = par_scores
 
-            DO_MUTATION = self.needs_mutation
-            for k in range(self.parents_count, self.population_size, 2):
-                r1, r2 = get_parents_inds()
-                pvar1 = par[r1]
-                pvar2 = par[r2]
-                
-                ch1, ch2 = self.crossover(pvar1, pvar2)
-                
-                if DO_MUTATION:
-                    ch1 = self.mut(ch1)
-                    ch2 = self.mut_middle(ch2, pvar1, pvar2)               
+            if self.fill_children is None:  # default fill children behaviour
+                DO_MUTATION = self.needs_mutation
+                for k in range(self.parents_count, self.population_size, 2):
 
-                pop[k] = ch1
-                pop[k+1] = ch2
+                    r1, r2 = get_parents_inds()
+
+                    pvar1 = pop[r1]  # equal to par[r1], but better for cache
+                    pvar2 = pop[r2]
+
+                    ch1, ch2 = self.crossover(pvar1, pvar2)
+
+                    if DO_MUTATION:
+                        ch1 = self.mut(ch1)
+                        ch2 = self.mut_middle(ch2, pvar1, pvar2)
+
+                    pop[k] = ch1
+                    pop[k+1] = ch2
+            else:  # custom behaviour
+                self.fill_children(pop, self.parents_count)
 
             if apply_function_to_parents:
                 scores = self.set_function(pop)
@@ -992,7 +1025,7 @@ class geneticalgorithm2:
     #endregion
 
     #region MUTATION
-    def mut(self, x: np.ndarray):
+    def mut(self, x: array1D):
         """
         just mutation
         """
@@ -1007,7 +1040,7 @@ class geneticalgorithm2:
             
         return x
 
-    def mut_middle(self, x: np.ndarray, p1: np.ndarray, p2: np.ndarray):
+    def mut_middle(self, x: array1D, p1: array1D, p2: array1D):
         """
         mutation oriented on parents
         """
@@ -1036,18 +1069,21 @@ class geneticalgorithm2:
 
     #region Set functions
 
+
     @staticmethod
-    def default_set_function(function_for_set: Callable[[np.ndarray], float]):
+    def default_set_function(function_for_set: Callable[[array1D], float]):
         """
         simple function for creating set_function 
         function_for_set just applyes to each row of population
         """
-        def func(matrix: np.ndarray):
-            return np.array([function_for_set(matrix[i]) for i in range(matrix.shape[0])])
+        def func(matrix: array2D):
+            return np.array(
+                [function_for_set(row) for row in matrix]
+            )
         return func
 
     @staticmethod
-    def vectorized_set_function(function_for_set: Callable[[np.ndarray], float]):
+    def vectorized_set_function(function_for_set: Callable[[array1D], float]):
         """
         works like default, but faster for big populations and slower for little
         function_for_set just applyes to each row of population
@@ -1057,12 +1093,13 @@ class geneticalgorithm2:
         return func
 
     @staticmethod
-    def set_function_multiprocess(function_for_set: Callable[[np.ndarray], float], n_jobs: int = -1):
+    def set_function_multiprocess(function_for_set: Callable[[array1D], float], n_jobs: int = -1):
         """
         like function_for_set but uses joblib with n_jobs (-1 goes to count of available processors)
         """
         from joblib import Parallel, delayed
-        def func(matrix: np.ndarray):
+
+        def func(matrix: array2D):
             result = Parallel(n_jobs=n_jobs)(delayed(function_for_set)(matrix[i]) for i in range(matrix.shape[0]))
             return np.array(result)
         return func
